//******************************************************************************
//******************************************************************************
//*   Copyright (C) 2011 K. Vasudevan. Indian Institute of Technology Kanpur.
//*   All Rights Reserved.
//******************************************************************************
// This code uses both prefix and preamble. Channel estimation is done using
// the preamble. Data detection is done using the estimated channel
// coefficients. Synchronization is done using the preamble.
// Frequency offset introduced 26/05/2011.
// Convolutional coding introduced 6/12/2011.
// Preamble amplitude adjustment (factor of 0.5) done on 02/04/2012.
// This code uses genlib() which generates binary files and is hence faster
// than dmt10. All if-else statements have been removed. ML channel estimate
// and non-ideal timing is assumed.
// The demodulation-matched filtering operation does not result in aliasing.
// Normalize_Alpha_Beta code incorporated in Dec1_Process and Dec2_Process.
// Postamble introduced on 01/08/2012.
// Faster? version of dmt14_e.
// Code tested on SCILAB ver 5.3.3.
//******************************************************************************
//clear;
//getd(pwd());                    // Get all functions in the current directory.
timer();
//******************************************************************************
// Constants.
//******************************************************************************
MAX_FRAMES=100;                      // No. of OFDM symbols to be simulated.
PRE_AMB_SIZE=512;                                     // Length of the preamble.
OFDM_SIZE=4096;           // Length of OFDM symbol, excluding the cyclic prefix.
POST_AMB_SIZE=128;                         // Length of postamble, must be even.
BUF_SIZE=4;                                                 // Length of buffer.
DATA_SIZE=OFDM_SIZE-round(POST_AMB_SIZE/2)-BUF_SIZE;          // Length of data.
PD_SIZE=2*DATA_SIZE+POST_AMB_SIZE;             // Length of postamble plus data.
PRE_AMB_QPSK_AMP=sqrt(0.5*PRE_AMB_SIZE/OFDM_SIZE); // Amp. of preamble QPSK sym.
CHAN_MEM=9;                                         // Length of channel memory.
CHAN_LEN=CHAN_MEM+1;                                       // Length of channel.
EST_CHAN_MEM=2*CHAN_MEM;                     // Estimated memory of the channel.
EST_CHAN_LEN=EST_CHAN_MEM+1;                        // Estimated channel length.
FRAME_SIZE=2*OFDM_SIZE+PRE_AMB_SIZE+EST_CHAN_MEM;                 // Frame size.
POST_RED_SIZE=POST_AMB_SIZE-2*BUF_SIZE;            // Size of reduced postamble.
IP_FAC=8;                                     // Interpolation factor at the rx.
IP_FAC_ZP=2*IP_FAC-2;                                 // Zero padding at the rx.
//******************************************************************************
// Frequency offset parameters.
//******************************************************************************
FOFF_VEC_SIZE=FRAME_SIZE+CHAN_LEN-1;                // Freq. offset vector size.
FOFF=0.04;           // Freq. offset lies in the interval [-0.04, 0.04] radians.
FOFF_EST_RX_SIZE=PRE_AMB_SIZE+EST_CHAN_LEN-1; // Size of rx. vec. for freq. est.
MAX_FOFF=0.04;                                            // Max . freq. offset.
MAX_FOFF2=0.005;                               // Max. freq. offset in 2nd step.
ML_FOFF1_BINS=64;                    // No. of freq. bins in 1st step of ML est.
ML_FOFF2_BINS=64;                    // No. of freq. bins in 2nd step of ML est.
//******************************************************************************
// SNR parameters.
//******************************************************************************
SNR_DB=8.0;                                                       // SNR in dB.
FADE_MEAN=0.0;                                                // Mean of fading.
FADE_VAR=0.5;                               // Variance per dimension of fading.
FADE_STD=sqrt(FADE_VAR);                            // Std. deviation of fading.
NOISE_MEAN=0.0;                                                 // Mean of AWGN.
NOISE_VAR=CHAN_LEN/(OFDM_SIZE*10^(0.1*SNR_DB));    // Var per dimension of AWGN.
NOISE_STD=sqrt(NOISE_VAR);                           // Std. deviation of noise.
//******************************************************************************
// QPSK symbols. Real part (data) maps to MSB. Imag. part (parity) maps to
// LSB.
//******************************************************************************
CONSTELL_SIZE=4;
QPSK_Map(1)=  1+%i;
QPSK_Map(2)=  1-%i;
QPSK_Map(3)= -1+%i;
QPSK_Map(4)= -1-%i;
for i=1:CONSTELL_SIZE
 QPSK_SYM(i,:)=QPSK_Map(i)*ones(1,2*DATA_SIZE);
end
//******************************************************************************
// Generate preamble QPSK symbols.
//******************************************************************************
[Re_Pre_Amb_Sym,Im_Pre_Amb_Sym,Pre_Amb_Sym]=Get_QPSK_Sym(PRE_AMB_SIZE);
Pre_Amb_Sym=Pre_Amb_Sym*PRE_AMB_QPSK_AMP;
Pre_Amb_IFFT=fft(Pre_Amb_Sym.',1).'; // IFFT. There is division by PRE_AMB_SIZE.
//******************************************************************************
// Generate postamble QPSK symbols.
//******************************************************************************
[Re_Post_Amb_Sym,Im_Post_Amb_Sym,Post_Amb_Sym]=Get_QPSK_Sym(POST_AMB_SIZE);
//******************************************************************************
// Generate buffer QPSK symbols.
//******************************************************************************
[Re_Buf_Sym,Im_Buf_Sym,Buf_Sym]=Get_QPSK_Sym(BUF_SIZE);
//******************************************************************************
// Get channel input matrix and channel estimation matrix.
//******************************************************************************
Chan_Ip_Matrix=zeros(PRE_AMB_SIZE-EST_CHAN_MEM,EST_CHAN_LEN);     // Initialize.
for i=1:PRE_AMB_SIZE-EST_CHAN_MEM
 Chan_Ip_Matrix(i,:)=Pre_Amb_IFFT(EST_CHAN_LEN+i-1:-1:i);
end
//******************************************************************************
// Least squares channel estimation matrix.
//******************************************************************************
 Chan_Est_Matrix=inv(Chan_Ip_Matrix'*Chan_Ip_Matrix)*Chan_Ip_Matrix';
//******************************************************************************
// Get the matched filter.
//******************************************************************************
MF_Coeff=conj([Pre_Amb_IFFT(PRE_AMB_SIZE:-1:1)]);             // Matched filter.
MF_Energy=MF_Coeff*MF_Coeff';                   // Energy of the matched filter.
MF_Coeff=MF_Coeff/MF_Energy;                       // Normalized to unit energy.
RX_MF_SIZE=FOFF_EST_RX_SIZE;                                  // Size of rx. MF.
//******************************************************************************
// Get the frequency offset matrix for a fixed frequency offset.
//******************************************************************************
//FOFF_Vec=0:FOFF_VEC_SIZE-1;
//FOFF_Vec=FOFF_Vec*%i*FOFF;
//FOFF_Vec=exp(FOFF_Vec);
//******************************************************************************
// Frequency offset estimation matrix for step 1.
//******************************************************************************
Freq_Step=2*MAX_FOFF/ML_FOFF1_BINS;
y=[0:ML_FOFF1_BINS]';
y=y*Freq_Step-MAX_FOFF;         // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix=y*x;
FOFF_Est_Matrix=exp(-%i*FOFF_Est_Matrix);
clear x;
clear y;
//******************************************************************************
// Frequency offset estimation matrix for step 2.
//******************************************************************************
Freq_Step2=2*MAX_FOFF2/ML_FOFF2_BINS;
y2=[0:ML_FOFF2_BINS]';
y2=y2*Freq_Step2-MAX_FOFF2;     // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix2=y2*x;
FOFF_Est_Matrix2=exp(-%i*FOFF_Est_Matrix2);
clear x;
clear y2;
//******************************************************************************
// Frequency offset estimation error measurements.
//******************************************************************************
global RMS_Err_FOFF1;
global RMS_Err_FOFF2;
global Max_Err_FOFF1;
global Max_Err_FOFF2;
RMS_Err_FOFF1=0.0;         // RMS frequency offset estimation error in 1st step.
Max_Err_FOFF1=-1.0;    // Maximum frequency offset estimation error in 1st step.
RMS_Err_FOFF2=0.0;         // RMS frequency offset estimation error in 2nd step.
Max_Err_FOFF2=-1.0;    // Maximum frequency offset estimation error in 2nd step.
//******************************************************************************
// Generator matrix for the rate-1/2 systematic convolutional encoder:
// [1 (1+D+D^2+D^3+D^4)/(1+D^3+D^4)]. Refer to Digital Communications and
// Signal Processing by K Vasudevan, Universities Press, for the notation
// followed.
//******************************************************************************
NC=[1 0 1];                          // Numerator coefficients in reverse order.
DC=[1 1 1];                        // Denominator coefficients in reverse order.
NP=poly(NC,'D','coeff');                                // Numerator polynomial.
DP=poly(DC,'D','coeff');                              // Denominator polynomial.
RSC_Gen_Vec=modulo(abs(ldiv(NP,DP,DATA_SIZE)),2);     // Column vector of coeff.
NC_Size=size(NC);
NUM_INPUTS=2;                                      // No. of input combinations.
NUM_STATES=2^(NC_Size(2)-1);                           // No. of trellis states.
RSC_MEM=NC_Size(2)-1;                                 // No. of memory elements.
[N_State,P_State,P_Ip,Op,Ga_Inx,Gb_Inx]=Get_Trellis();
//******************************************************************************
// Get the interleaver and deinterleaver map for the turbo code.
//******************************************************************************
[Iv_Map,Div_Map]=Get_Iv_Div_Map(DATA_SIZE);
//******************************************************************************
// Get the interleaver and deinterleaver map for OFDM.
//******************************************************************************
[OFDM_Iv_Map,OFDM_Div_Map]=Get_OFDM_Iv_Div_Map(PD_SIZE);
//******************************************************************************
// Cyclic prefix (CP) indices.
//******************************************************************************
CP_ST_INX=2*OFDM_SIZE-EST_CHAN_MEM+1;       // Start index for extraction of CP.
CP_ED_INX=2*OFDM_SIZE;                        // End index for extraction of CP.
//******************************************************************************
// Starting and ending indices for the postamble in the data frame.
//******************************************************************************
POST_START=2*DATA_SIZE+1;
POST_END=PD_SIZE;
//******************************************************************************
// Interleaved indices of the postamble.
//******************************************************************************
Div_Post_Inx=OFDM_Div_Map(POST_START:POST_END);
//******************************************************************************
// Starting and ending indices for interleaved information symbols in the
// data frame excluding the buffer and postamble.
//******************************************************************************
GINX1=DATA_SIZE+1;
GINX2=2*DATA_SIZE;
//******************************************************************************
// Initialize gammas.
//******************************************************************************
global My_Gamma1;
global My_Gamma2;
My_Gamma1=zeros(CONSTELL_SIZE,DATA_SIZE);
My_Gamma2=zeros(CONSTELL_SIZE,DATA_SIZE);
//******************************************************************************
global Ber;
global C_Ber;
global Erasure;
Ber=0;                                          // Uncoded bit-error-rate (BER).
C_Ber=0;                                            // BER after turbo decoding.
Erasure=0;                                          // Number of frame erasures.
//******************************************************************************
Simulate();
//******************************************************************************
RMS_Err_FOFF1=sqrt(RMS_Err_FOFF1/MAX_FRAMES);
RMS_Err_FOFF2=sqrt(RMS_Err_FOFF2/(MAX_FRAMES-Erasure));
timer()
//******************************************************************************
