//******************************************************************************
//******************************************************************************
//*   Copyright (C) 2015 K. Vasudevan. Indian Institute of Technology Kanpur.
//*   All Rights Reserved.
//******************************************************************************
// This code uses both prefix and preamble. Channel estimation is done using
// the preamble. Data detection is done using the estimated channel
// coefficients. Synchronization is done using the preamble.
// Frequency offset introduced 26/05/2011.
// Convolutional coding introduced 6/12/2011.
// Preamble amplitude adjustment (factor of 0.5) done on 02/04/2012.
// This code uses genlib() which generates binary files.
// All if-else statements have been removed. ML channel estimate
// and non-ideal timing is assumed.
// The demodulation-matched filtering operation does not result in aliasing.
// Normalize_Alpha_Beta code incorporated in Dec1_Process and Dec2_Process.
// Postamble introduced on 01/08/2012.
// Robust turbo coding (normalizing exponents) introduced on 12/12/2012.
// Rate increased to 1 on 19/02/2013.
// 2nd order receiver diversity introduced on 31/03/2013. Identical
// frequency offset; independent channel and noise on both diversity arms.
// 10th June 2015: 2nd order transmit diversity introduced.
// 24th July 2015: 3rd attempt to estimate and cancel the frequency offset is
//                 introduced.
// 28th Aug 2015: The random number generator is initialized using the date.
//                The "Ber" and "C_Ber" results are written onto a file.
// 27th June 2016: The size of the channel input matrix is now
//                 L_p x L_{hr}. This is a correction over dmt31_e.sce.
// 16th July 2016: Turbo decoding using a super trellis for both decoders.
//                 This is an improvement over dmt33_e.sce.
// "ldiv" is not used for generating the impulse response of the
// constituent convolutional codes.
// Code tested on SCILAB ver 5.5.2.
//******************************************************************************
//clear;
//getd(pwd());                    // Get all functions in the current directory.
timer();
//******************************************************************************
// Add a word of caution.
//******************************************************************************
mprintf('Running this code will append to the data files\n');
mprintf('u_ber.txt and c_ber.txt.\n');
mprintf('Is this ok?.\n');
mprintf('\n');
pause;
//******************************************************************************
// Initialize the random number generators with the date.
//******************************************************************************
n=getdate("s");
grand("setsd",n);
//******************************************************************************
// Open files for appending.
//******************************************************************************
SNR_DB=6.5;                                               // SNR per bit in dB.
x=string(SNR_DB);
y = strsubst(x, '.', '_');
fd_u_ber_w = mopen(pwd()+'/u_ber_'+y+'.txt', 'at');         // Uncoded BER file.
fd_c_ber_w = mopen(pwd()+'/c_ber_'+y+'.txt', 'at');           // Coded BER file.
fd_f_eer_w = mopen(pwd()+'/f_eer_'+y+'.txt', 'at');             // Erasure file.
//******************************************************************************
// Constants.
//******************************************************************************
MAX_FRAMES=1;                      // No. of OFDM symbols to be simulated.
PRE_AMB_SIZE=512;                                    // Length of the preamble.
OFDM_SIZE=4096;           // Length of OFDM symbol, excluding the cyclic prefix.
POST_AMB_SIZE=256;                         // Length of postamble, must be even.
BUF_SIZE=4;                 // Length of buffer before and after the data frame.
DATA_SIZE=OFDM_SIZE-POST_AMB_SIZE-2*BUF_SIZE;   // Length of data, must be even.
PD_SIZE=DATA_SIZE+POST_AMB_SIZE;               // Length of postamble plus data.
PRE_AMB_QPSK_AMP=sqrt(2*PRE_AMB_SIZE/OFDM_SIZE);   // Amp. of preamble QPSK sym.
WELL_HEIGHT=0.0000;                 // Height of the well for the two preambles.
CHAN_MEM=9;                                         // Length of channel memory.
CHAN_LEN=CHAN_MEM+1;                                       // Length of channel.
EST_CHAN_MEM=2*CHAN_MEM;                     // Estimated memory of the channel.
EST_CHAN_LEN=EST_CHAN_MEM+1;                        // Estimated channel length.
FRAME_SIZE=OFDM_SIZE+PRE_AMB_SIZE+2*EST_CHAN_MEM;                 // Frame size.
IP_FAC=16;                                    // Interpolation factor at the rx.
IP_FAC_ZP=IP_FAC-1;                                   // Zero padding at the rx.
//******************************************************************************
// Frequency offset parameters.
//******************************************************************************
FOFF_VEC_SIZE=FRAME_SIZE+CHAN_LEN-1;                // Freq. offset vector size.
FOFF=0.04;           // Freq. offset lies in the interval [-0.04, 0.04] radians.
FOFF_EST_RX_SIZE=PRE_AMB_SIZE+EST_CHAN_LEN-1; // Size of rx. vec. for freq. est.
MAX_FOFF=0.04;                                            // Max . freq. offset.
MAX_FOFF2=0.005;                               // Max. freq. offset in 2nd step.
ML_FOFF1_BINS=64;                    // No. of freq. bins in 1st step of ML est.
ML_FOFF2_BINS=64;                    // No. of freq. bins in 2nd step of ML est.
//******************************************************************************
// SNR parameters.
//******************************************************************************
FADE_MEAN=0.0;                                                // Mean of fading.
FADE_VAR=0.5;                               // Variance per dimension of fading.
FADE_STD=sqrt(FADE_VAR);                            // Std. deviation of fading.
NOISE_MEAN=0.0;                                                 // Mean of AWGN.
NOISE_VAR=4.0*CHAN_LEN/(OFDM_SIZE*10^(0.1*SNR_DB));      // Var per dim of AWGN.
NOISE_STD=sqrt(NOISE_VAR);                           // Std. deviation of noise.
//******************************************************************************
// QPSK symbols. Real part (data) maps to MSB. Imag. part (parity) maps to
// LSB.
//******************************************************************************
CONSTELL_SIZE=4;
QPSK_Map(1)=  1+%i;
QPSK_Map(2)=  1-%i;
QPSK_Map(3)= -1+%i;
QPSK_Map(4)= -1-%i;
for i=1:CONSTELL_SIZE
 QPSK_SYM(i,:)=QPSK_Map(i)*ones(1,DATA_SIZE);
end
//******************************************************************************
// Get the interleaver and deinterleaver map for the preamble.
//******************************************************************************
[Iv_Map_Pre_Amb,Div_Map_Pre_Amb]=Get_Iv_Div_Map(PRE_AMB_SIZE);
//******************************************************************************
// Generate preamble QPSK symbols for 1st transmitter.
//******************************************************************************
Pre_Amb_Sym1=Get_QPSK_Sym(PRE_AMB_SIZE);
Pre_Amb_Sym1=Pre_Amb_Sym1*WELL_HEIGHT;
Tmp_Pre_Amb=Get_QPSK_Sym(PRE_AMB_SIZE/2);
Pre_Amb_Sym1(Iv_Map_Pre_Amb(1:PRE_AMB_SIZE/2))=Tmp_Pre_Amb;
Pre_Amb_Sym1=Pre_Amb_Sym1*PRE_AMB_QPSK_AMP;
Pre_Amb_IFFT1=fft(Pre_Amb_Sym1.',1).';     // There is division by PRE_AMB_SIZE.
//******************************************************************************
// Generate preamble QPSK symbols for 2nd transmitter.
//******************************************************************************
Pre_Amb_Sym2=Get_QPSK_Sym(PRE_AMB_SIZE);
Pre_Amb_Sym2=Pre_Amb_Sym2*WELL_HEIGHT;
Tmp_Pre_Amb=Get_QPSK_Sym(PRE_AMB_SIZE/2);
Pre_Amb_Sym2(Iv_Map_Pre_Amb(PRE_AMB_SIZE/2+1:$))=Tmp_Pre_Amb;
Pre_Amb_Sym2=Pre_Amb_Sym2*PRE_AMB_QPSK_AMP;
Pre_Amb_IFFT2=fft(Pre_Amb_Sym2.',1).';     // There is division by PRE_AMB_SIZE.
//******************************************************************************
// Clear the variables.
//******************************************************************************
clear Pre_Amb_Sym1;
clear Pre_Amb_Sym2;
clear Tmp_Pre_Amb;
clear Iv_Map_Pre_Amb;
clear Div_Map_Pre_Amb;
//******************************************************************************
// Generate postamble QPSK symbols for 1st transmitter.
//******************************************************************************
[Post_Amb_Sym1]=Get_QPSK_Sym(POST_AMB_SIZE);
//******************************************************************************
// Generate postamble QPSK symbols for 2nd transmitter.
//******************************************************************************
[Post_Amb_Sym2]=Get_QPSK_Sym(POST_AMB_SIZE);
//******************************************************************************
// Generate buffer QPSK symbols.
//******************************************************************************
[Buf_Sym1]=Get_QPSK_Sym(BUF_SIZE);                              // Start buffer.
[Buf_Sym2]=Get_QPSK_Sym(BUF_SIZE);                                // End buffer.
//******************************************************************************
// Get channel input matrix and channel estimation matrix for 1st transmitter.
//******************************************************************************
Chan_Ip_Matrix1=zeros(PRE_AMB_SIZE,EST_CHAN_LEN);                 // Initialize.
Pre_Amb_IFFT_Suf=[Pre_Amb_IFFT1 Pre_Amb_IFFT1(1:EST_CHAN_MEM)];
for i=1:PRE_AMB_SIZE
 Chan_Ip_Matrix1(i,:)=Pre_Amb_IFFT_Suf(EST_CHAN_LEN+i-1:-1:i);
end
clear Pre_Amb_IFFT_Suf;
//******************************************************************************
// Get channel input matrix and channel estimation matrix for 2nd transmitter.
//******************************************************************************
Chan_Ip_Matrix2=zeros(PRE_AMB_SIZE,EST_CHAN_LEN);                 // Initialize.
Pre_Amb_IFFT_Suf=[Pre_Amb_IFFT2 Pre_Amb_IFFT2(1:EST_CHAN_MEM)];
for i=1:PRE_AMB_SIZE
 Chan_Ip_Matrix2(i,:)=Pre_Amb_IFFT_Suf(EST_CHAN_LEN+i-1:-1:i);
end
clear Pre_Amb_IFFT_Suf;
//******************************************************************************
// Least squares channel estimation matrix for 1st transmitter.
//******************************************************************************
 Chan_Est_Matrix1=inv(Chan_Ip_Matrix1'*Chan_Ip_Matrix1)*Chan_Ip_Matrix1';
//******************************************************************************
// Least squares channel estimation matrix for 2nd transmitter.
//******************************************************************************
 Chan_Est_Matrix2=inv(Chan_Ip_Matrix2'*Chan_Ip_Matrix2)*Chan_Ip_Matrix2';
//******************************************************************************
// Length of the filter matched to: preamble convolved with channel.
//******************************************************************************
RX_MF_SIZE=FOFF_EST_RX_SIZE;                                  // Size of rx. MF.
//******************************************************************************
// Get the matched filter for the preamble from 1st transmitter.
//******************************************************************************
MF_Coeff1=conj([Pre_Amb_IFFT1(PRE_AMB_SIZE:-1:1)]);           // Matched filter.
MF_Energy=MF_Coeff1*MF_Coeff1';                 // Energy of the matched filter.
MF_Coeff1=MF_Coeff1/MF_Energy;                     // Normalized to unit energy.
//******************************************************************************
// Get the matched filter for the preamble from 2nd transmitter.
//******************************************************************************
MF_Coeff2=conj([Pre_Amb_IFFT2(PRE_AMB_SIZE:-1:1)]);           // Matched filter.
MF_Energy=MF_Coeff2*MF_Coeff2';                 // Energy of the matched filter.
MF_Coeff2=MF_Coeff2/MF_Energy;                     // Normalized to unit energy.
//******************************************************************************
// Get the frequency offset matrix for a fixed frequency offset.
//******************************************************************************
//FOFF_Vec=0:FOFF_VEC_SIZE-1;
//FOFF_Vec=FOFF_Vec*%i*FOFF;
//FOFF_Vec=exp(FOFF_Vec);
//******************************************************************************
// Frequency offset estimation matrix for step 1.
//******************************************************************************
Freq_Step=2*MAX_FOFF/ML_FOFF1_BINS;
y=[0:ML_FOFF1_BINS]';
y=y*Freq_Step-MAX_FOFF;         // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix=y*x;
FOFF_Est_Matrix=exp(-%i*FOFF_Est_Matrix);
clear x;
clear y;
//******************************************************************************
// Frequency offset estimation matrix for step 2.
//******************************************************************************
Freq_Step2=2*MAX_FOFF2/ML_FOFF2_BINS;
y2=[0:ML_FOFF2_BINS]';
y2=y2*Freq_Step2-MAX_FOFF2;     // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix2=y2*x;
FOFF_Est_Matrix2=exp(-%i*FOFF_Est_Matrix2);
clear x;
clear y2;
//******************************************************************************
// Frequency offset estimation error measurements.
//******************************************************************************
RMS_Err_FOFF1=0.0;         // RMS frequency offset estimation error in 1st step.
RMS_Err_FOFF2=0.0;         // RMS frequency offset estimation error in 2nd step.
RMS_Err_FOFF3=0.0;         // RMS frequency offset estimation error in 3rd step.
Max_Err_FOFF1= -1.0;   // Maximum frequency offset estimation error in 1st step.
Max_Err_FOFF2= -1.0;   // Maximum frequency offset estimation error in 2nd step.
Max_Err_FOFF3= -1.0;   // Maximum frequency offset estimation error in 3rd step.
//******************************************************************************
// Generator matrix for the rate-1/2 systematic convolutional encoder:
// [1 (1+D+D^2+D^3+D^4)/(1+D^3+D^4)]. Refer to Digital Communications and
// Signal Processing by K Vasudevan, Universities Press, for the notation
// followed.
//******************************************************************************
NC=[1 0 1];                                           // Numerator coefficients.
DC=[1 1 1];                                         // Denominator coefficients.
NC_Size=size(NC);
NUM_INPUTS=2;                                      // No. of input combinations.
NUM_STATES=2^(NC_Size(2)-1);                           // No. of trellis states.
RSC_MEM=NC_Size(2)-1;                                 // No. of memory elements.
[N_State,P_State,P_Ip,Op,Ga_Inx,Gb_Inx]=Get_Trellis();
RSC_Gen_Vec=Get_RSC_Gen_Vec();                          // Get impulse response.
//******************************************************************************
//                           Get the super trellis.
//******************************************************************************
MOD_NUM_INPUTS=4;
NUM_S_STATES=16;
MOD_CONSTELL_SIZE=16;
[N_S_State,P_S_State,P_S_Ip,Ga_S_Inx,Gb_S_Inx]=Get_S_Trellis();
//******************************************************************************
// Get the interleaver and deinterleaver map for the turbo code.
//******************************************************************************
[Iv_Map,Div_Map]=Get_Iv_Div_Map(DATA_SIZE);
//******************************************************************************
// Get the interleaver and deinterleaver map for OFDM.
//******************************************************************************
[OFDM_Iv_Map,OFDM_Div_Map]=Get_OFDM_Iv_Div_Map(PD_SIZE);
//******************************************************************************
// Cyclic prefix (CP) indices.
//******************************************************************************
CP_ST_INX=OFDM_SIZE-EST_CHAN_MEM+1;         // Start index for extraction of CP.
CP_ED_INX=OFDM_SIZE;                          // End index for extraction of CP.
//******************************************************************************
// Starting and ending indices for the postamble in the data frame.
//******************************************************************************
POST_START=DATA_SIZE+1;
POST_END=PD_SIZE;
//******************************************************************************
// Interleaved indices of the postamble.
//******************************************************************************
Iv_Post_Inx=OFDM_Iv_Map(POST_START:POST_END);
//******************************************************************************
// Resizing of the deinterleaved FFT output.
//******************************************************************************
//Div_Rx_FFT_Op=zeros(IP_FAC,2*OFDM_SIZE);
//******************************************************************************
// Starting and ending indices for the data (consisting of information symbols
// and interleaved information symbols).
//******************************************************************************
//DATA_START=BUF_SIZE+1;
//DATA_END=BUF_SIZE+2*DATA_SIZE;
//******************************************************************************
// Starting and ending indices for information symbols in the data frame.
//******************************************************************************
//INFO_START=BUF_SIZE+1;
//INFO_END=INFO_START+DATA_SIZE-1;
//******************************************************************************
// Starting and ending indices for interleaved information symbols in the
// data frame excluding the buffer and postamble.
//******************************************************************************
GINX1=(DATA_SIZE/2)+1;
GINX2=DATA_SIZE;
//******************************************************************************
// Minimum exponent.
//******************************************************************************
MIN_EXPO= -30.0;
//******************************************************************************
// Initialize gammas.
//******************************************************************************
My_Gamma1=zeros(MOD_CONSTELL_SIZE,DATA_SIZE);
My_Gamma2=zeros(MOD_CONSTELL_SIZE,DATA_SIZE);
//******************************************************************************
Ber=0;                                          // Uncoded bit-error-rate (BER).
C_Ber=0;                                            // BER after turbo decoding.
Erasure=0;                                          // Number of frame erasures.
//******************************************************************************
// Simulation.
//******************************************************************************
for Frame_Cnt=1:MAX_FRAMES
//******************************************************************************
// Get the random frequency offset and the frequency offset matrix.
//******************************************************************************
 FOFF=2.0*MAX_FOFF*(grand(1,1,'def')-0.5);
// FOFF=0.0;
 FOFF_Vec=0:FOFF_VEC_SIZE-1;
 FOFF_Vec=FOFF_Vec*%i*FOFF;
 FOFF_Vec=exp(FOFF_Vec);
//******************************************************************************
// Generate coded QPSK symbols and append postamble for 1st and 2nd
// transmitter.
//******************************************************************************
 [C1,C2,My_Data]=Get_Coded_QPSK();
 C1_Post=[C1 Post_Amb_Sym1];                                 // 1st transmitter.
 C2_Post=[C2 Post_Amb_Sym2];                                 // 2nd transmitter.
//******************************************************************************
// Interleaving and add buffer symbols for 1st transmitter.
//******************************************************************************
 Tx_Sym1=[Buf_Sym1 C1_Post(OFDM_Div_Map(:)) Buf_Sym2];
//******************************************************************************
// Interleaving and add buffer symbols for 2nd transmitter.
//******************************************************************************
 Tx_Sym2=[Buf_Sym1 C2_Post(OFDM_Div_Map(:)) Buf_Sym2];
//******************************************************************************
 Sym_IFFT1=fft(Tx_Sym1.',1).';        // 1st tx. There is division by OFDM_SIZE.
 Sym_IFFT2=fft(Tx_Sym2.',1).';        // 2nd tx. There is division by OFDM_SIZE.
//******************************************************************************
// Add cyclic prefix, cyclic postfix and preamble for 1st transmitter.
//******************************************************************************
 Chan_Ip1=[Pre_Amb_IFFT1 Pre_Amb_IFFT1(1:EST_CHAN_MEM)...
           Sym_IFFT1(CP_ST_INX:CP_ED_INX) Sym_IFFT1];
//******************************************************************************
// Add cyclic prefix, cyclic postfix and preamble for 2nd transmitter.
//******************************************************************************
 Chan_Ip2=[Pre_Amb_IFFT2 Pre_Amb_IFFT2(1:EST_CHAN_MEM)...
           Sym_IFFT2(CP_ST_INX:CP_ED_INX) Sym_IFFT2];
//******************************************************************************
// Get channel for 1st tx, 1st rx (h11).
//******************************************************************************
 Chan11=Get_Channel(CHAN_LEN,FADE_MEAN,FADE_STD);
 Chan_Ext_Zeros=[Chan11 zeros(1,OFDM_SIZE-CHAN_LEN)];           // Zero padding.
 Chan11_FFT=fft(Chan_Ext_Zeros.',-1).';                  // OFDM_SIZE-point FFT.
//******************************************************************************
// Get channel for 2nd tx, 1st rx (h21).
//******************************************************************************
 Chan21=Get_Channel(CHAN_LEN,FADE_MEAN,FADE_STD);
 Chan_Ext_Zeros=[Chan21 zeros(1,OFDM_SIZE-CHAN_LEN)];           // Zero padding.
 Chan21_FFT=fft(Chan_Ext_Zeros.',-1).';                  // OFDM_SIZE-point FFT.
//******************************************************************************
// Get channel for 1st tx, 2nd rx (h12).
//******************************************************************************
 Chan12=Get_Channel(CHAN_LEN,FADE_MEAN,FADE_STD);
 Chan_Ext_Zeros=[Chan12 zeros(1,OFDM_SIZE-CHAN_LEN)];           // Zero padding.
 Chan12_FFT=fft(Chan_Ext_Zeros.',-1).';                  // OFDM_SIZE-point FFT.
//******************************************************************************
// Get channel for 2nd tx, 2nd rx (h22).
//******************************************************************************
 Chan22=Get_Channel(CHAN_LEN,FADE_MEAN,FADE_STD);
 Chan_Ext_Zeros=[Chan22 zeros(1,OFDM_SIZE-CHAN_LEN)];           // Zero padding.
 Chan22_FFT=fft(Chan_Ext_Zeros.',-1).';                  // OFDM_SIZE-point FFT.
//******************************************************************************
// Get channel output for 1st rx (diversity arm).
//******************************************************************************
 Chan_Op1=Get_Channel_Op(Chan11,Chan21);
 FOFF_Est_Rx1=Chan_Op1(1:FOFF_EST_RX_SIZE);
 FOFF_Est_Rx_Mat1=repmat(FOFF_Est_Rx1.',1,ML_FOFF1_BINS+1);
 FOFF_Est_Rx_Mat1=FOFF_Est_Rx_Mat1.*(FOFF_Est_Matrix.');
//******************************************************************************
// Get channel output for 2nd rx (diversity arm).
//******************************************************************************
 Chan_Op2=Get_Channel_Op(Chan12,Chan22);
 FOFF_Est_Rx2=Chan_Op2(1:FOFF_EST_RX_SIZE);
 FOFF_Est_Rx_Mat2=repmat(FOFF_Est_Rx2.',1,ML_FOFF1_BINS+1);
 FOFF_Est_Rx_Mat2=FOFF_Est_Rx_Mat2.*(FOFF_Est_Matrix.');
//******************************************************************************
// ML frequency offset estimation, step 1.
// This is for the 1st diversity arm (rx) and preamble from 1st tx.
//******************************************************************************
 FOFF_Sync_Op1=convol(MF_Coeff1,FOFF_Est_Rx_Mat1(:));
 [Max_Sync_Op,Max_Inx]=max(abs(FOFF_Sync_Op1));
 Est_FOFF_Inx=int((Max_Inx-1)/FOFF_EST_RX_SIZE);
 Est_FOFF1_11=Est_FOFF_Inx*Freq_Step-MAX_FOFF;
 Diff_Inx11=Max_Inx-PRE_AMB_SIZE-Est_FOFF_Inx*FOFF_EST_RX_SIZE;
//******************************************************************************
// ML frequency offset estimation, step 1.
// This is for the 1st diversity arm (rx) and preamble from 2nd tx.
//******************************************************************************
 FOFF_Sync_Op1=convol(MF_Coeff2,FOFF_Est_Rx_Mat1(:));
 [Max_Sync_Op,Max_Inx]=max(abs(FOFF_Sync_Op1));
 Est_FOFF_Inx=int((Max_Inx-1)/FOFF_EST_RX_SIZE);
 Est_FOFF1_21=Est_FOFF_Inx*Freq_Step-MAX_FOFF;
 Diff_Inx21=Max_Inx-PRE_AMB_SIZE-Est_FOFF_Inx*FOFF_EST_RX_SIZE;
//******************************************************************************
// ML frequency offset estimation, step 1.
// This is for the 2nd diversity arm (rx) and preamble from 1st tx.
//******************************************************************************
 FOFF_Sync_Op1=convol(MF_Coeff1,FOFF_Est_Rx_Mat2(:));
 [Max_Sync_Op,Max_Inx]=max(abs(FOFF_Sync_Op1));
 Est_FOFF_Inx=int((Max_Inx-1)/FOFF_EST_RX_SIZE);
 Est_FOFF1_12=Est_FOFF_Inx*Freq_Step-MAX_FOFF;
 Diff_Inx12=Max_Inx-PRE_AMB_SIZE-Est_FOFF_Inx*FOFF_EST_RX_SIZE;
//******************************************************************************
// ML frequency offset estimation, step 1.
// This is for the 2nd diversity arm (rx) and preamble from 2nd tx.
//******************************************************************************
 FOFF_Sync_Op1=convol(MF_Coeff2,FOFF_Est_Rx_Mat2(:));
 [Max_Sync_Op,Max_Inx]=max(abs(FOFF_Sync_Op1));
 Est_FOFF_Inx=int((Max_Inx-1)/FOFF_EST_RX_SIZE);
 Est_FOFF1_22=Est_FOFF_Inx*Freq_Step-MAX_FOFF;
 Diff_Inx22=Max_Inx-PRE_AMB_SIZE-Est_FOFF_Inx*FOFF_EST_RX_SIZE;
//******************************************************************************
// Average frequency offset estimate.
//******************************************************************************
 Avg_Est_FOFF1=0.25*(Est_FOFF1_11+Est_FOFF1_21+Est_FOFF1_12+Est_FOFF1_22);
 Avg_Freq_Est_Err1=FOFF-Avg_Est_FOFF1;
 RMS_Err_FOFF1=RMS_Err_FOFF1+Avg_Freq_Est_Err1*Avg_Freq_Est_Err1;
 Max_Err_FOFF1=max([Max_Err_FOFF1 abs(Avg_Freq_Est_Err1)]);
 Avg_Est_FOFF_Vec=exp(-%i*[0:FOFF_EST_RX_SIZE-1]*Avg_Est_FOFF1);
//******************************************************************************
// Frame detection 1st attempt in the presence of frequency offset.
//******************************************************************************
 if ((Diff_Inx11 >= 0) & (Diff_Inx11 < CHAN_LEN) &...
     (Diff_Inx21 >= 0) & (Diff_Inx21 < CHAN_LEN) &...
     (Diff_Inx12 >= 0) & (Diff_Inx12 < CHAN_LEN) &...
     (Diff_Inx22 >= 0) & (Diff_Inx22 < CHAN_LEN)) then
//******************************************************************************
// Get starting and ending indices for channel estimation.
//******************************************************************************
  Start_Inx=CHAN_LEN+Diff_Inx11;
  End_Inx=Start_Inx+PRE_AMB_SIZE-1;
//******************************************************************************
// Channel estimation 1st attempt. This is for the 1st diversity arm (rx)
// and preamble from 1st tx.
//******************************************************************************
  Demod_Rx1=Avg_Est_FOFF_Vec.*FOFF_Est_Rx1;
  Rx_Sig=Demod_Rx1(Start_Inx:End_Inx);
  Chan11_Est=Chan_Est_Matrix1*Rx_Sig.';
  Chan_Est_Ext_Zeros=[Chan11_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan11_FFT=fft(Chan_Est_Ext_Zeros,-1).';
//******************************************************************************
// Plot the actual and estimated channel magnitude response (h11).
//******************************************************************************
//  set(gca(),"auto_clear","off");               // Two plots on the same graph.
//  a=gca();                                                // Get current axes.
//  a.box='on';                                            // Put plot in a box.
//  plot2d(abs(Chan11_FFT),style=5);
//  plot2d(abs(Est_Chan11_FFT),style=2);
//  legends(["Actual";"Estimated"],[[5],[2]], with_box=%t, opt="?");
//******************************************************************************
// Get rx. MF by convolving the estimated h11 with the preamble from 1st tx.
//******************************************************************************
  Rx_MF11=convol(Chan11_Est,Pre_Amb_IFFT1);
  Rx_MF11=conj([Rx_MF11(RX_MF_SIZE:-1:1)]);     // Filter matched to rx. signal.
  Rx_MF_Energy=Rx_MF11*Rx_MF11';                // Energy of the matched filter.
  Rx_MF11=Rx_MF11/Rx_MF_Energy;                    // Normalized to unit energy.
//  pause;
//******************************************************************************
// Channel estimation 1st attempt. This is for the 1st diversity arm (rx)
// and preamble from 2nd tx.
//******************************************************************************
  Chan21_Est=Chan_Est_Matrix2*Rx_Sig.';
  Chan_Est_Ext_Zeros=[Chan21_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan21_FFT=fft(Chan_Est_Ext_Zeros,-1).';
//******************************************************************************
// Plot the actual and estimated channel magnitude response (h21).
//******************************************************************************
//  set(gca(),"auto_clear","off");               // Two plots on the same graph.
//  a=gca();                                                // Get current axes.
//  a.box='on';                                            // Put plot in a box.
//  plot2d(abs(Chan21_FFT),style=5);
//  plot2d(abs(Est_Chan21_FFT),style=2);
//  legends(["Actual";"Estimated"],[[5],[2]], with_box=%t, opt="?");
//******************************************************************************
// Get rx. MF by convolving the estimated h21 with the preamble from 2nd tx.
//******************************************************************************
  Rx_MF21=convol(Chan21_Est,Pre_Amb_IFFT2);
  Rx_MF21=conj([Rx_MF21(RX_MF_SIZE:-1:1)]);     // Filter matched to rx. signal.
  Rx_MF_Energy=Rx_MF21*Rx_MF21';                // Energy of the matched filter.
  Rx_MF21=Rx_MF21/Rx_MF_Energy;                    // Normalized to unit energy.
//  pause;
//******************************************************************************
// Channel estimation 1st attempt. This is for the 2nd diversity arm (rx)
// and preamble from 1st tx.
//******************************************************************************
  Demod_Rx2=Avg_Est_FOFF_Vec.*FOFF_Est_Rx2;
  Rx_Sig=Demod_Rx2(Start_Inx:End_Inx);
  Chan12_Est=Chan_Est_Matrix1*Rx_Sig.';
  Chan_Est_Ext_Zeros=[Chan12_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan12_FFT=fft(Chan_Est_Ext_Zeros,-1).';
//******************************************************************************
// Plot the actual and estimated channel magnitude response (h12).
//******************************************************************************
//  set(gca(),"auto_clear","off");               // Two plots on the same graph.
//  a=gca();                                                // Get current axes.
//  a.box='on';                                            // Put plot in a box.
//  plot2d(abs(Chan12_FFT),style=5);
//  plot2d(abs(Est_Chan12_FFT),style=2);
//  legends(["Actual";"Estimated"],[[5],[2]], with_box=%t, opt="?");
//******************************************************************************
// Get rx. MF by convolving the estimated h12 with the preamble from 1st tx.
//******************************************************************************
  Rx_MF12=convol(Chan12_Est,Pre_Amb_IFFT1);
  Rx_MF12=conj([Rx_MF12(RX_MF_SIZE:-1:1)]);        // Filter matched to rx. sig.
  Rx_MF_Energy=Rx_MF12*Rx_MF12';                // Energy of the matched filter.
  Rx_MF12=Rx_MF12/Rx_MF_Energy;                    // Normalized to unit energy.
//  pause;
//******************************************************************************
// Channel estimation 1st attempt. This is for the 2nd diversity arm (rx)
// and preamble from 2nd tx.
//******************************************************************************
  Chan22_Est=Chan_Est_Matrix2*Rx_Sig.';
  Chan_Est_Ext_Zeros=[Chan22_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan22_FFT=fft(Chan_Est_Ext_Zeros,-1).';
//******************************************************************************
// Plot the actual and estimated channel magnitude response (h22).
//******************************************************************************
//  set(gca(),"auto_clear","off");               // Two plots on the same graph.
//  a=gca();                                                // Get current axes.
//  a.box='on';                                            // Put plot in a box.
//  plot2d(abs(Chan22_FFT),style=5);
//  plot2d(abs(Est_Chan22_FFT),style=2);
//  legends(["Actual";"Estimated"],[[5],[2]], with_box=%t, opt="?");
//******************************************************************************
// Get rx. MF by convolving the estimated h22 with the preamble from 2nd tx.
//******************************************************************************
  Rx_MF22=convol(Chan22_Est,Pre_Amb_IFFT2);
  Rx_MF22=conj([Rx_MF22(RX_MF_SIZE:-1:1)]);        // Filter matched to rx. sig.
  Rx_MF_Energy=Rx_MF22*Rx_MF22';                // Energy of the matched filter.
  Rx_MF22=Rx_MF22/Rx_MF_Energy;                    // Normalized to unit energy.
//  pause;
//******************************************************************************
// Get the demodulated rx. sig. matrix. 1st diversity arm. Frequency
// offset estimation 2nd attempt.
//******************************************************************************
  FOFF_Est_Rx_Mat1_2=repmat(Demod_Rx1.',1,ML_FOFF2_BINS+1);
  FOFF_Est_Rx_Mat1_2=FOFF_Est_Rx_Mat1_2.*(FOFF_Est_Matrix2.');
//******************************************************************************
// Get the demodulated rx. sig. matrix. 2nd diversity arm. Frequency
// offset estimation 2nd attempt.
//******************************************************************************
  FOFF_Est_Rx_Mat2_2=repmat(Demod_Rx2.',1,ML_FOFF2_BINS+1);
  FOFF_Est_Rx_Mat2_2=FOFF_Est_Rx_Mat2_2.*(FOFF_Est_Matrix2.');
//******************************************************************************
// ML frequency offset estimation, step 2.
// MF obtained by convolving estimated h11 with the preamble from 1st tx.
//******************************************************************************
  FOFF_Sync_Op2=convol(Rx_MF11,FOFF_Est_Rx_Mat1_2(:));
  [Max_Sync_Op,Max_Inx]=max(abs(FOFF_Sync_Op2));
  Est_FOFF_Inx=int((Max_Inx-1)/FOFF_EST_RX_SIZE);
  Est_FOFF2_11=Est_FOFF_Inx*Freq_Step2-MAX_FOFF2;
  Freq_Est_Err=Avg_Est_FOFF1+Est_FOFF2_11-FOFF;
//  pause;
//******************************************************************************
// ML frequency offset estimation, step 2.
// MF obtained by convolving estimated h21 with the preamble from 2nd tx.
//******************************************************************************
  FOFF_Sync_Op2=convol(Rx_MF21,FOFF_Est_Rx_Mat1_2(:));
  [Max_Sync_Op,Max_Inx]=max(abs(FOFF_Sync_Op2));
  Est_FOFF_Inx=int((Max_Inx-1)/FOFF_EST_RX_SIZE);
  Est_FOFF2_21=Est_FOFF_Inx*Freq_Step2-MAX_FOFF2;
  Freq_Est_Err=Avg_Est_FOFF1+Est_FOFF2_21-FOFF;
//  pause;
//******************************************************************************
// ML frequency offset estimation, step 2.
// MF obtained by convolving estimated h12 with the preamble from 1st tx.
//******************************************************************************
  FOFF_Sync_Op2=convol(Rx_MF12,FOFF_Est_Rx_Mat2_2(:));
  [Max_Sync_Op,Max_Inx]=max(abs(FOFF_Sync_Op2));
  Est_FOFF_Inx=int((Max_Inx-1)/FOFF_EST_RX_SIZE);
  Est_FOFF2_12=Est_FOFF_Inx*Freq_Step2-MAX_FOFF2;
  Freq_Est_Err=Avg_Est_FOFF1+Est_FOFF2_12-FOFF;
//  pause;
//******************************************************************************
// ML frequency offset estimation, step 2.
// MF obtained by convolving estimated h22 with the preamble from 2nd tx.
//******************************************************************************
  FOFF_Sync_Op2=convol(Rx_MF22,FOFF_Est_Rx_Mat2_2(:));
  [Max_Sync_Op,Max_Inx]=max(abs(FOFF_Sync_Op2));
  Est_FOFF_Inx=int((Max_Inx-1)/FOFF_EST_RX_SIZE);
  Est_FOFF2_22=Est_FOFF_Inx*Freq_Step2-MAX_FOFF2;
  Freq_Est_Err=Avg_Est_FOFF1+Est_FOFF2_22-FOFF;
//  pause;
//******************************************************************************
// Average frequency offset estimate in the 2nd step.
//******************************************************************************
  Avg_Est_FOFF2=0.25*(Est_FOFF2_11+Est_FOFF2_21+Est_FOFF2_12+Est_FOFF2_22);
  Avg_Freq_Est_Err2=FOFF-(Avg_Est_FOFF1+Avg_Est_FOFF2);
  RMS_Err_FOFF2=RMS_Err_FOFF2+Avg_Freq_Est_Err2*Avg_Freq_Est_Err2;
  Max_Err_FOFF2=max([Max_Err_FOFF2 abs(Avg_Freq_Est_Err2)]);
  Avg_Est_FOFF_Vec=exp(-%i*[0:FOFF_VEC_SIZE-1]*(Avg_Est_FOFF1+Avg_Est_FOFF2));
//******************************************************************************
// Demodulation using the 2nd estimate of the frequency offset.
// 1st diversity arm.
//******************************************************************************
  Chan_Op1=Chan_Op1.*Avg_Est_FOFF_Vec;
//******************************************************************************
// Demodulation using the 2nd estimate of the frequency offset.
// 2nd diversity arm.
//******************************************************************************
  Chan_Op2=Chan_Op2.*Avg_Est_FOFF_Vec;
//******************************************************************************
// Final channel estimation (h11). 1st diversity arm (rx).
//******************************************************************************
  Rx_Sig=Chan_Op1(Start_Inx:End_Inx);
  Chan11_Est=Chan_Est_Matrix1*Rx_Sig.';
  Chan_Est_Ext_Zeros=[Chan11_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan11_FFT=fft(Chan_Est_Ext_Zeros,-1).';           // Data size chan. FFT.
//******************************************************************************
// Plot the actual and estimated channel magnitude response (h11).
//******************************************************************************
//  subplot(221);
//  set(gca(),"auto_clear","off");               // Two plots on the same graph.
//  a=gca();                                                // Get current axes.
//  a.box='on';                                            // Put plot in a box.
//  plot2d(abs(Chan11_FFT),style=5);
//  plot2d(abs(Est_Chan11_FFT),style=2);
//  legends(["Actual";"Estimated"],[[5],[2]], with_box=%t, opt="?");
//  pause;
//******************************************************************************
// Final channel estimation (h21). 1st diversity arm (rx).
//******************************************************************************
  Chan21_Est=Chan_Est_Matrix2*Rx_Sig.';
  Chan_Est_Ext_Zeros=[Chan21_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan21_FFT=fft(Chan_Est_Ext_Zeros,-1).';           // Data size chan. FFT.
//******************************************************************************
// Plot the actual and estimated channel magnitude response (h21).
//******************************************************************************
//  subplot(222);
//  set(gca(),"auto_clear","off");               // Two plots on the same graph.
//  a=gca();                                                // Get current axes.
//  a.box='on';                                            // Put plot in a box.
//  plot2d(abs(Chan21_FFT),style=5);
//  plot2d(abs(Est_Chan21_FFT),style=2);
//  legends(["Actual";"Estimated"],[[5],[2]], with_box=%t, opt="?");
//  pause;
//******************************************************************************
// Final channel estimation (h12). 2nd diversity arm (rx).
//******************************************************************************
  Rx_Sig=Chan_Op2(Start_Inx:End_Inx);
  Chan12=Chan_Est_Matrix1*Rx_Sig.';
  Chan_Est_Ext_Zeros=[Chan12;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan12_FFT=fft(Chan_Est_Ext_Zeros,-1).';           // Data size chan. FFT.
//******************************************************************************
// Plot the actual and estimated channel magnitude response (h12).
//******************************************************************************
//  subplot(223);
//  set(gca(),"auto_clear","off");               // Two plots on the same graph.
//  a=gca();                                                // Get current axes.
//  a.box='on';                                            // Put plot in a box.
//  plot2d(abs(Chan12_FFT),style=5);
//  plot2d(abs(Est_Chan12_FFT),style=2);
//  legends(["Actual";"Estimated"],[[5],[2]], with_box=%t, opt="?");
//  pause;
//******************************************************************************
// Final channel estimation (h22). 2nd diversity arm (rx).
//******************************************************************************
  Chan22=Chan_Est_Matrix2*Rx_Sig.';
  Chan_Est_Ext_Zeros=[Chan22;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan22_FFT=fft(Chan_Est_Ext_Zeros,-1).';           // Data size chan. FFT.
//******************************************************************************
// Plot the actual and estimated channel magnitude response (h22).
//******************************************************************************
//  subplot(224);
//  set(gca(),"auto_clear","off");               // Two plots on the same graph.
//  a=gca();                                                // Get current axes.
//  a.box='on';                                            // Put plot in a box.
//  plot2d(abs(Chan22_FFT),style=5);
//  plot2d(abs(Est_Chan22_FFT),style=2);
//  legends(["Actual";"Estimated"],[[5],[2]], with_box=%t, opt="?");
//  pause;
//******************************************************************************
// Interpolated FFT using zero-padding. 1st diversity arm.
//******************************************************************************
  Start_Inx2=Start_Inx+PRE_AMB_SIZE+EST_CHAN_MEM; // Include the postfix length.
  End_Inx2=Start_Inx2+OFDM_SIZE-1;
  Rx_Sig2=Chan_Op1(Start_Inx2:End_Inx2);                             // For FFT.
  Rx_Sig2_1=[Rx_Sig2 zeros(1,IP_FAC_ZP*OFDM_SIZE)];
  Rx_FFT_Op1=fft(Rx_Sig2_1.',-1).';                  // FFT of received samples.
//******************************************************************************
// Interpolated FFT using zero-padding. 2nd diversity arm.
//******************************************************************************
  Rx_Sig2=Chan_Op2(Start_Inx2:End_Inx2);                             // For FFT.
  Rx_Sig2_1=[Rx_Sig2 zeros(1,IP_FAC_ZP*OFDM_SIZE)];
  Rx_FFT_Op2=fft(Rx_Sig2_1.',-1).';                  // FFT of received samples.
//******************************************************************************
// Get the noise free postamble part of the FFT output using the estimated
// channel and obtain the matched filter. 1st diversity arm.
//******************************************************************************
  Post_FFT_Op=Post_Amb_Sym1.*Est_Chan11_FFT(BUF_SIZE+Iv_Post_Inx)+...
              Post_Amb_Sym2.*Est_Chan21_FFT(BUF_SIZE+Iv_Post_Inx);
  OFDM_Post_FFT_Op=zeros(1,OFDM_SIZE);
  OFDM_Post_FFT_Op(BUF_SIZE+Iv_Post_Inx)=Post_FFT_Op;
  Up_Post_FFT_Op=zeros(1,IP_FAC*OFDM_SIZE);
  Up_Post_FFT_Op(1:IP_FAC:$)=OFDM_Post_FFT_Op;
  Up_Post_FFT_Op_MF1=conj(Up_Post_FFT_Op($:-1:1));
//******************************************************************************
// Get the noise free postamble part of the FFT output using the estimated
// channel and obtain the matched filter. 2nd diversity arm.
//******************************************************************************
  Post_FFT_Op=Post_Amb_Sym1.*Est_Chan12_FFT(BUF_SIZE+Iv_Post_Inx)+...
              Post_Amb_Sym2.*Est_Chan22_FFT(BUF_SIZE+Iv_Post_Inx);
  OFDM_Post_FFT_Op=zeros(1,OFDM_SIZE);
  OFDM_Post_FFT_Op(BUF_SIZE+Iv_Post_Inx)=Post_FFT_Op;
  Up_Post_FFT_Op=zeros(1,IP_FAC*OFDM_SIZE);
  Up_Post_FFT_Op(1:IP_FAC:$)=OFDM_Post_FFT_Op;
  Up_Post_FFT_Op_MF2=conj(Up_Post_FFT_Op($:-1:1));
//******************************************************************************
// Postamble matched filtering. 1st diversity arm.
//******************************************************************************
  Post_Corr_Op=convol(Up_Post_FFT_Op_MF1,Rx_FFT_Op1);
  [Post_Corr_Max,Post_Corr_Max_Inx]=max(abs(Post_Corr_Op));
  Diff_Inx2_1=Post_Corr_Max_Inx-IP_FAC*OFDM_SIZE;
//  pause;
//******************************************************************************
// Postamble matched filtering. 2nd diversity arm.
//******************************************************************************
  Post_Corr_Op=convol(Up_Post_FFT_Op_MF2,Rx_FFT_Op2);
  [Post_Corr_Max,Post_Corr_Max_Inx]=max(abs(Post_Corr_Op));
  Diff_Inx2_2=Post_Corr_Max_Inx-IP_FAC*OFDM_SIZE;
//  pause;
//******************************************************************************
// Average frequency offset estimate, 3rd attempt. This could be used for
// uplink transmissions.
//******************************************************************************
  Avg_Est_FOFF3=0.5*(Diff_Inx2_1+Diff_Inx2_2)*2*%pi/(OFDM_SIZE*IP_FAC);
  Avg_Freq_Est_Err3=FOFF-(Avg_Est_FOFF1+Avg_Est_FOFF2+Avg_Est_FOFF3);
  RMS_Err_FOFF3=RMS_Err_FOFF3+Avg_Freq_Est_Err3*Avg_Freq_Est_Err3;
  Max_Err_FOFF3=max([Max_Err_FOFF3 abs(Avg_Freq_Est_Err3)]);
  Avg_Est_FOFF_Vec=exp(-%i*[0:FOFF_VEC_SIZE-1]*Avg_Est_FOFF3);
//  pause;
//******************************************************************************
// Demodulation using the 3rd estimate of the frequency offset.
// 1st diversity arm.
//******************************************************************************
  Chan_Op1=Chan_Op1.*Avg_Est_FOFF_Vec;
//******************************************************************************
// Demodulation using the 3rd estimate of the frequency offset.
// 2nd diversity arm.
//******************************************************************************
  Chan_Op2=Chan_Op2.*Avg_Est_FOFF_Vec;
//******************************************************************************
// Data detection, 1st diversity arm.
//******************************************************************************
  Rx_Sig2=Chan_Op1(Start_Inx2:End_Inx2);                             // For FFT.
  Rx_FFT_Op2_1=fft(Rx_Sig2.',-1).';                  // FFT of received samples.
//******************************************************************************
// Data detection, 2nd diversity arm.
//******************************************************************************
  Rx_Sig2=Chan_Op2(Start_Inx2:End_Inx2);                             // For FFT.
  Rx_FFT_Op2_2=fft(Rx_Sig2.',-1).';                  // FFT of received samples.
//******************************************************************************
// Get the data part of the deinterleaved FFT output and the data part of the
// deinterleaved FFT of the channel estimate. 1st diversity arm.
//******************************************************************************
  Rx_PD_Start=BUF_SIZE+1;
  Rx_PD_End=Rx_PD_Start+PD_SIZE-1;
  Rx_FFT_Op1_Data=Rx_FFT_Op2_1(Rx_PD_Start:Rx_PD_End);
  Iv_Rx_FFT_Op1_Data=Rx_FFT_Op1_Data(OFDM_Iv_Map(1:DATA_SIZE));
  Iv_Est_Chan11_FFT=Est_Chan11_FFT(BUF_SIZE+OFDM_Iv_Map(1:DATA_SIZE));
  Iv_Est_Chan21_FFT=Est_Chan21_FFT(BUF_SIZE+OFDM_Iv_Map(1:DATA_SIZE));
//******************************************************************************
// Get the data part of the deinterleaved FFT output and the data part of the
// deinterleaved FFT of the channel estimate. 2nd diversity arm.
//******************************************************************************
  Rx_FFT_Op1_Data=Rx_FFT_Op2_2(Rx_PD_Start:Rx_PD_End);
  Iv_Rx_FFT_Op2_Data=Rx_FFT_Op1_Data(OFDM_Iv_Map(1:DATA_SIZE));
  Iv_Est_Chan12_FFT=Est_Chan12_FFT(BUF_SIZE+OFDM_Iv_Map(1:DATA_SIZE));
  Iv_Est_Chan22_FFT=Est_Chan22_FFT(BUF_SIZE+OFDM_Iv_Map(1:DATA_SIZE));
//******************************************************************************
// Compute noise variance in two dimensions, before the FFT operation. 1st
// diversity arm.
//******************************************************************************
  Rx_Sig=Chan_Op1(Start_Inx:End_Inx);
  N_Vec=Rx_Sig.'-Chan_Ip_Matrix1*Chan11_Est-Chan_Ip_Matrix2*Chan21_Est;
  Est_N_Var1=N_Vec'*N_Vec/PRE_AMB_SIZE;
//******************************************************************************
//******************************************************************************
// Compute noise variance in two dimensions, before the FFT operation. 2nd
// diversity arm.
//******************************************************************************
  Rx_Sig=Chan_Op2(Start_Inx:End_Inx);
  N_Vec=Rx_Sig.'-Chan_Ip_Matrix1*Chan12_Est-Chan_Ip_Matrix2*Chan22_Est;
  Est_N_Var2=N_Vec'*N_Vec/PRE_AMB_SIZE;
//******************************************************************************
// Compute average noise variance.
//******************************************************************************
  Est_N_Var=0.5*(Est_N_Var1+Est_N_Var2);
//******************************************************************************
// Compute squared Euclidean distances.
//******************************************************************************
  for tx1_sym=1:CONSTELL_SIZE
   for tx2_sym=1:CONSTELL_SIZE
    Dist(tx1_sym,tx2_sym,:)=abs(Iv_Rx_FFT_Op1_Data...
                           -QPSK_SYM(tx1_sym,:).*Iv_Est_Chan11_FFT...
                           -QPSK_SYM(tx2_sym,:).*Iv_Est_Chan21_FFT).^2 ...
                           +abs(Iv_Rx_FFT_Op2_Data...
                           -QPSK_SYM(tx1_sym,:).*Iv_Est_Chan12_FFT...
                           -QPSK_SYM(tx2_sym,:).*Iv_Est_Chan22_FFT).^2;
    Dist1((tx1_sym-1)*CONSTELL_SIZE+tx2_sym,:)=abs(Iv_Rx_FFT_Op1_Data...
                           -QPSK_SYM(tx1_sym,:).*Iv_Est_Chan11_FFT...
                           -QPSK_SYM(tx2_sym,:).*Iv_Est_Chan21_FFT).^2;
    Dist2((tx1_sym-1)*CONSTELL_SIZE+tx2_sym,:)=abs(Iv_Rx_FFT_Op2_Data...
                           -QPSK_SYM(tx1_sym,:).*Iv_Est_Chan12_FFT...
                           -QPSK_SYM(tx2_sym,:).*Iv_Est_Chan22_FFT).^2;
   end
  end
//******************************************************************************
  [M1,K1]=min(Dist(:,1,:),Dist(:,2,:),Dist(:,3,:),Dist(:,4,:));    // Min. dist.
  [M2,K2]=min(Dist(1,:,:),Dist(2,:,:),Dist(3,:,:),Dist(4,:,:));    // Min. dist.
  M3=matrix(M1,CONSTELL_SIZE,DATA_SIZE);
  M4=matrix(M2,CONSTELL_SIZE,DATA_SIZE);
  [M5,K5]=min(M3(1,:),M3(2,:),M3(3,:),M3(4,:));
  Rx_Sym=QPSK_Map(K5);                                         // Column vector.
  Ber=Ber+nnz(real(C1-Rx_Sym.'));                                // Uncoded BER.
//  pause;
//******************************************************************************
// Compute gammas.
//******************************************************************************
  My_Exp1= -Dist1(:,:)/(OFDM_SIZE*real(Est_N_Var));
  My_Exp2= -Dist2(:,:)/(OFDM_SIZE*real(Est_N_Var));
//  pause;
  My_Exp1=My_Exp1-repmat(max(My_Exp1,'r'),MOD_CONSTELL_SIZE,1);
  My_Exp2=My_Exp2-repmat(max(My_Exp2,'r'),MOD_CONSTELL_SIZE,1);
  Exp1_Inx=find(My_Exp1 < MIN_EXPO);
  Exp2_Inx=find(My_Exp2 < MIN_EXPO);
  My_Exp1(Exp1_Inx)= MIN_EXPO;
  My_Exp2(Exp2_Inx)= MIN_EXPO;
//  pause;
  My_Gamma1(:,:)=exp(My_Exp1);
  My_Gamma2(:,:)=exp(My_Exp2);
//******************************************************************************
// The MAP (BCJR) algorithm.
//******************************************************************************
  Prob=0.25*ones(MOD_NUM_INPUTS,DATA_SIZE);
//******************************************************************************
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process(Prob);
  [Prob]=Dec1_Process(Prob);
  [Prob]=Dec2_Process_End(Prob);
//******************************************************************************
  tmp_p_tx1=zeros(NUM_INPUTS,DATA_SIZE);
  tmp_p_tx1(1,:)=Prob(1,:)+Prob(2,:)+Prob(1,Iv_Map(:))+Prob(3,Iv_Map(:));
  tmp_p_tx1(2,:)=Prob(3,:)+Prob(4,:)+Prob(2,Iv_Map(:))+Prob(4,Iv_Map(:));
//******************************************************************************
  [M6,K6]=max(tmp_p_tx1(1,:),tmp_p_tx1(2,:));         // Max. prob. and indices.
  Rx_Sym1=1-2*(K6-1);
  C_Ber=C_Ber+nnz(My_Data-Rx_Sym1);
//******************************************************************************
 else
  Erasure=Erasure+1;
 end
end
//******************************************************************************
RMS_Err_FOFF1=sqrt(RMS_Err_FOFF1/MAX_FRAMES);
RMS_Err_FOFF2=sqrt(RMS_Err_FOFF2/(MAX_FRAMES-Erasure));
RMS_Err_FOFF3=sqrt(RMS_Err_FOFF3/(MAX_FRAMES-Erasure));
F_U_Ber=Ber/((MAX_FRAMES-Erasure)*DATA_SIZE);         // Fractional uncoded BER.
F_C_Ber=C_Ber/((MAX_FRAMES-Erasure)*DATA_SIZE);          //Fractional coded BER.
F_F_Eer=Erasure/MAX_FRAMES;              // Fractional frame erasure error rate.
mfprintf(fd_u_ber_w,'%f %.10f\n',SNR_DB,F_U_Ber);
mfprintf(fd_c_ber_w,'%f %.10f\n',SNR_DB,F_C_Ber);
mfprintf(fd_f_eer_w,'%f %.10f\n',SNR_DB,F_F_Eer);
mclose(fd_u_ber_w);
mclose(fd_c_ber_w);
mclose(fd_f_eer_w);
timer()
//******************************************************************************
