//******************************************************************************
//******************************************************************************
//*   Copyright (C) 2010 K. Vasudevan. Indian Institute of Technology Kanpur.
//*   All Rights Reserved.
//******************************************************************************
// This code uses both prefix and preamble. Channel estimation is done using
// the preamble. Data detection is done using the estimated channel
// coefficients. Synchronization is done using the preamble.
// Code tested on SCILAB ver 5.3.3.
//******************************************************************************
clear;
tic();
timer();
//******************************************************************************
// Constants.
//******************************************************************************
MAX_FRAMES=10000;                      // No. of OFDM symbols to be simulated.
PRE_AMB_SIZE=512;                                     // Length of the preamble.
OFDM_SIZE=4096;           // Length of OFDM symbol, excluding the cyclic prefix.
PRE_AMB_QPSK_AMP=sqrt(PRE_AMB_SIZE/OFDM_SIZE);     // Amp. of preamble QPSK sym.
CHAN_MEM=9;                                         // Length of channel memory.
CHAN_LEN=CHAN_MEM+1;                                       // Length of channel.
EST_CHAN_MEM=2*CHAN_MEM;                     // Estimated memory of the channel.
EST_CHAN_LEN=EST_CHAN_MEM+1;                        // Estimated channel length.
FRAME_SIZE=OFDM_SIZE+PRE_AMB_SIZE+EST_CHAN_MEM;                   // Frame size.
//******************************************************************************
// SNR parameters.
//******************************************************************************
SNR_DB=40.0;                                                       // SNR in dB.
FADE_MEAN=0.0;                                                // Mean of fading.
FADE_VAR=0.5;                               // Variance per dimension of fading.
FADE_STD=sqrt(FADE_VAR);                            // Std. deviation of fading.
NOISE_MEAN=0.0;                                                 // Mean of AWGN.
NOISE_VAR=CHAN_LEN/(OFDM_SIZE*10^(0.1*SNR_DB));    // Var per dimension of AWGN.
NOISE_STD=sqrt(NOISE_VAR);                           // Std. deviation of noise.
//******************************************************************************
// Channel estimation. 0: estimated channel; 1: ideal channel.
//******************************************************************************
IDEAL_CHEST=1;
//******************************************************************************
// MMSE channel estimation. 0: least squares; 1: MMSE.
//******************************************************************************
MMSE_CHEST=0;
//******************************************************************************
// QPSK symbols.
//******************************************************************************
QPSK_SYM1=(+1+%i)*ones(1,OFDM_SIZE);                  // QPSK symbols of type 1.
QPSK_SYM2=(-1+%i)*ones(1,OFDM_SIZE);                  // QPSK symbols of type 2.
QPSK_SYM3=(-1-%i)*ones(1,OFDM_SIZE);                  // QPSK symbols of type 3.
QPSK_SYM4=(+1-%i)*ones(1,OFDM_SIZE);                  // QPSK symbols of type 4.
//******************************************************************************
// Generate preamble QPSK symbols.
//******************************************************************************
Re_Pre_Amb_Bits=grand(1,PRE_AMB_SIZE,'def');
Im_Pre_Amb_Bits=grand(1,PRE_AMB_SIZE,'def');
Re_Pre_Amb_Bits=int(2*Re_Pre_Amb_Bits);
Im_Pre_Amb_Bits=int(2*Im_Pre_Amb_Bits);
Re_Pre_Amb_Sym=1.0-2.0*Re_Pre_Amb_Bits;
Im_Pre_Amb_Sym=1.0-2.0*Im_Pre_Amb_Bits;
Pre_Amb_Sym=(Re_Pre_Amb_Sym+%i*Im_Pre_Amb_Sym)*PRE_AMB_QPSK_AMP;
Pre_Amb_IFFT=fft(Pre_Amb_Sym.',1).'; // IFFT. There is division by PRE_AMB_SIZE.
//******************************************************************************
// Get channel input matrix and channel estimation matrix.
//******************************************************************************
Chan_Ip_Matrix=zeros(PRE_AMB_SIZE-EST_CHAN_MEM,EST_CHAN_LEN);     // Initialize.
for i=1:PRE_AMB_SIZE-EST_CHAN_MEM
 Chan_Ip_Matrix(i,:)=Pre_Amb_IFFT(EST_CHAN_LEN+i-1:-1:i);
end
//******************************************************************************
// Least squares channel estimation matrix.
//******************************************************************************
if MMSE_CHEST == 0 then
 Chan_Est_Matrix=inv(Chan_Ip_Matrix'*Chan_Ip_Matrix)*Chan_Ip_Matrix';
//******************************************************************************
// MMSE channel estimation matrix.
//******************************************************************************
elseif MMSE_CHEST == 1 then
 R1=Chan_Ip_Matrix*Chan_Ip_Matrix';
 R=2*FADE_VAR*R1+2*NOISE_VAR*eye(R1);
 MMSE_Chan_Est_Matrix=Chan_Ip_Matrix'*inv(R);
end
//******************************************************************************
// Get the matched filter.
//******************************************************************************
MF_Coeff=conj([Pre_Amb_IFFT(PRE_AMB_SIZE:-1:1)]);             // Matched filter.
MF_Energy=MF_Coeff*MF_Coeff';                   // Energy of the matched filter.
//******************************************************************************
Ber=0;                                                        // Bit-error-rate.
Erasure=0;                                   // Counts no. of frames in erasure.
//******************************************************************************
for Frame_Cnt=1:MAX_FRAMES
// clf();
//******************************************************************************
// Generate QPSK symbols.
//******************************************************************************
 Re_Tx_Bits=grand(1,OFDM_SIZE,'def');
 Im_Tx_Bits=grand(1,OFDM_SIZE,'def');
 Re_Tx_Bits=int(2*Re_Tx_Bits);
 Im_Tx_Bits=int(2*Im_Tx_Bits);
 Re_Tx_Sym=1.0-2.0*Re_Tx_Bits;
 Im_Tx_Sym=1.0-2.0*Im_Tx_Bits;
 Tx_Sym=Re_Tx_Sym+%i*Im_Tx_Sym;
//******************************************************************************
 Sym_IFFT=fft(Tx_Sym.',1).';            // IFFT. There is division by OFDM_SIZE.
//******************************************************************************
// Add cyclic prefix and preamble.
//******************************************************************************
 Chan_Ip=[Pre_Amb_IFFT Sym_IFFT(OFDM_SIZE-EST_CHAN_MEM+1:OFDM_SIZE) Sym_IFFT];
//******************************************************************************
// Get channel.
//******************************************************************************
 Re_Chan=zeros(1,CHAN_LEN);          // Initialize re. channel impulse response.
 Im_Chan=zeros(1,CHAN_LEN);          // Initialize im. channel impulse response.
 Re_Chan=grand(Re_Chan,'nor',FADE_MEAN,FADE_STD);                // Re. channel.
 Im_Chan=grand(Im_Chan,'nor',FADE_MEAN,FADE_STD);                // Im. channel.
 Chan=Re_Chan+%i*Im_Chan;                                    // Complex channel.
 Chan_Ext_Zeros=[Chan zeros(1,OFDM_SIZE-CHAN_LEN)];             // Zero padding.
 Chan_FFT=fft(Chan_Ext_Zeros.',-1).';                    // OFDM_SIZE-point FFT.
//******************************************************************************
 Chan_Op=convol(Chan,Chan_Ip);
 Re_Noise=grand(Chan_Op,'nor',NOISE_MEAN,NOISE_STD);
 Im_Noise=grand(Chan_Op,'nor',NOISE_MEAN,NOISE_STD);
 Noise=Re_Noise+%i*Im_Noise;
 Chan_Op=Chan_Op+Noise;
 Sync_Op=convol(MF_Coeff,Chan_Op)/MF_Energy;
 Avg_Sync_Op=sum(abs(Sync_Op))/size(Sync_Op,'c');
 [Peak_Sync_Op,Peak_Inx]=max(abs(Sync_Op));
 PAR(Frame_Cnt)=Peak_Sync_Op/Avg_Sync_Op;
// subplot(221);
// plot(abs(Sync_Op).^2);
// Rx_Sig=Chan_Op(EST_CHAN_LEN:FRAME_SIZE);             // Steady-state samples.
 if IDEAL_CHEST == 0 then
  Diff_Inx=Peak_Inx-PRE_AMB_SIZE;
 else
  Diff_Inx=0;
 end
 if ((Diff_Inx >= 0) & (Diff_Inx < CHAN_LEN)) then
  Start_Inx1=EST_CHAN_LEN-Diff_Inx;
  End_Inx1=PRE_AMB_SIZE+Start_Inx1-EST_CHAN_LEN;
  Rx_Sig1=Chan_Op(Start_Inx1:End_Inx1);
  if MMSE_CHEST == 0 then
   Chan_Est=Chan_Est_Matrix*Rx_Sig1.';
  elseif MMSE_CHEST == 1 then
   Chan_Est=MMSE_Chan_Est_Matrix*Rx_Sig1.';
  end
  Chan_Est_Ext_Zeros=[Chan_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan_FFT=fft(Chan_Est_Ext_Zeros,-1).';
//  subplot(222);
//  plot2d3(abs(Chan_Est));
//  subplot(224);
//  plot2d3(abs(Chan));
//  subplot(222);
//  plot(abs(Chan_FFT));
//  subplot(224);
//  plot(abs(Est_Chan_FFT))
  Start_Inx2=PRE_AMB_SIZE+EST_CHAN_LEN-Diff_Inx;
  End_Inx2=FRAME_SIZE-Diff_Inx;
  Rx_Sig2=Chan_Op(Start_Inx2:End_Inx2);                              // For FFT.
  Rx_FFT_Op=fft(Rx_Sig2.',-1).';                     // FFT of received samples.
  if IDEAL_CHEST == 0 then
   Dist1=abs(Rx_FFT_Op-QPSK_SYM1.*Est_Chan_FFT);         // Sq. dist. wrt sym 1.
   Dist2=abs(Rx_FFT_Op-QPSK_SYM2.*Est_Chan_FFT);         // Sq. dist. wrt sym 2.
   Dist3=abs(Rx_FFT_Op-QPSK_SYM3.*Est_Chan_FFT);         // Sq. dist. wrt sym 3.
   Dist4=abs(Rx_FFT_Op-QPSK_SYM4.*Est_Chan_FFT);         // Sq. dist. wrt sym 4.
  else
   Dist1=abs(Rx_FFT_Op-QPSK_SYM1.*Chan_FFT);             // Sq. dist. wrt sym 1.
   Dist2=abs(Rx_FFT_Op-QPSK_SYM2.*Chan_FFT);             // Sq. dist. wrt sym 2.
   Dist3=abs(Rx_FFT_Op-QPSK_SYM3.*Chan_FFT);             // Sq. dist. wrt sym 3.
   Dist4=abs(Rx_FFT_Op-QPSK_SYM4.*Chan_FFT);             // Sq. dist. wrt sym 4.
  end
  [M,K]=min(Dist1,Dist2,Dist3,Dist4);              // Get min. dist and indices.
  Rx_Sym=round(exp(%i*(K*%pi/2-%pi/4)));                   // Estimated symbols.
  Ber=Ber+nnz(Tx_Sym-Rx_Sym);
 else
  Erasure=Erasure+1;                                 // Erase the current frame.
 end
end
//******************************************************************************
timer()
toc()
//******************************************************************************
