//******************************************************************************
//******************************************************************************
//*   Copyright (C) 2011 K. Vasudevan. Indian Institute of Technology Kanpur.
//*   All Rights Reserved.
//******************************************************************************
// This code uses both prefix and preamble. Channel estimation is done using
// the preamble. Data detection is done using the estimated channel
// coefficients. Synchronization is done using the preamble.
// Frequency offset introduced 26/05/2011. 2-step frequency offset estimation.
// Code tested on SCILAB ver 5.3.2.
//******************************************************************************
clear;
timer();
getd(pwd());
//******************************************************************************
// Constants.
//******************************************************************************
MAX_FRAMES=10000;                       // No. of OFDM symbols to be simulated.
PRE_AMB_SIZE=512;                                     // Length of the preamble.
OFDM_SIZE=256;           // Length of OFDM symbol, excluding the cyclic prefix.
PRE_AMB_QPSK_AMP=sqrt(PRE_AMB_SIZE/OFDM_SIZE);     // Amp. of preamble QPSK sym.
CHAN_MEM=9;                                         // Length of channel memory.
CHAN_LEN=CHAN_MEM+1;                                       // Length of channel.
EST_CHAN_MEM=2*CHAN_LEN;                     // Estimated memory of the channel.
EST_CHAN_LEN=EST_CHAN_MEM+1;                        // Estimated channel length.
FRAME_SIZE=OFDM_SIZE+PRE_AMB_SIZE+EST_CHAN_MEM;                   // Frame size.
//******************************************************************************
// Frequency offset parameters.
//******************************************************************************
FOFF_VEC_SIZE=FRAME_SIZE+CHAN_LEN-1;                // Freq. offset vector size.
FOFF=0.04;           // Freq. offset lies in the interval [-0.04, 0.04] radians.
FOFF_EST_RX_SIZE=PRE_AMB_SIZE+EST_CHAN_LEN;   // Size of rx. vec. for freq. est.
MAX_FOFF=0.04;                                            // Max . freq. offset.
MAX_FOFF2=0.005;                               // Max. freq. offset in 2nd step.
ML_FOFF1_BINS=32;                    // No. of freq. bins in 1st step of ML est.
ML_FOFF2_BINS=32;                   // No. of freq. bins in 2nd step of ML est.
if abs(FOFF) > MAX_FOFF then
 halt('Freq offset exceeds maximum.');
end;
//******************************************************************************
// SNR parameters.
//******************************************************************************
SNR_DB=0.0;                                                       // SNR in dB.
FADE_MEAN=0.0;                                                // Mean of fading.
FADE_VAR=0.5;                               // Variance per dimension of fading.
FADE_STD=sqrt(FADE_VAR);                            // Std. deviation of fading.
NOISE_MEAN=0.0;                                                 // Mean of AWGN.
NOISE_VAR=CHAN_LEN/(OFDM_SIZE*10^(0.1*SNR_DB));    // Var per dimension of AWGN.
NOISE_STD=sqrt(NOISE_VAR);                           // Std. deviation of noise.
//******************************************************************************
// Channel estimation. 0: estimated channel; 1: ideal channel.
//******************************************************************************
IDEAL_CHEST=0;
//******************************************************************************
// MMSE channel estimation. 0: least squares; 1: MMSE.
//******************************************************************************
MMSE_CHEST=0;
//******************************************************************************
// QPSK symbols.
//******************************************************************************
QPSK_SYM1=(+1+%i)*ones(1,OFDM_SIZE);                  // QPSK symbols of type 1.
QPSK_SYM2=(-1+%i)*ones(1,OFDM_SIZE);                  // QPSK symbols of type 2.
QPSK_SYM3=(-1-%i)*ones(1,OFDM_SIZE);                  // QPSK symbols of type 3.
QPSK_SYM4=(+1-%i)*ones(1,OFDM_SIZE);                  // QPSK symbols of type 4.
//******************************************************************************
// Generate preamble QPSK symbols.
//******************************************************************************
[Re_Pre_Amb_Sym,Im_Pre_Amb_Sym,Pre_Amb_Sym]=Get_QPSK_Sym(PRE_AMB_SIZE);
Pre_Amb_Sym=Pre_Amb_Sym*PRE_AMB_QPSK_AMP;
Pre_Amb_IFFT=fft(Pre_Amb_Sym.',1).'; // IFFT. There is division by PRE_AMB_SIZE.
//******************************************************************************
// Get channel input matrix and channel estimation matrix.
//******************************************************************************
Chan_Ip_Matrix=zeros(PRE_AMB_SIZE-EST_CHAN_MEM,EST_CHAN_LEN);     // Initialize.
for i=1:PRE_AMB_SIZE-EST_CHAN_MEM
 Chan_Ip_Matrix(i,:)=Pre_Amb_IFFT(EST_CHAN_LEN+i-1:-1:i);
end
//******************************************************************************
// Least squares channel estimation matrix.
//******************************************************************************
if MMSE_CHEST == 0 then
 Chan_Est_Matrix=inv(Chan_Ip_Matrix'*Chan_Ip_Matrix)*Chan_Ip_Matrix';
//******************************************************************************
// MMSE channel estimation matrix.
//******************************************************************************
elseif MMSE_CHEST == 1 then
 R1=Chan_Ip_Matrix*Chan_Ip_Matrix';
 R=2*FADE_VAR*R1+2*NOISE_VAR*eye(R1);
 MMSE_Chan_Est_Matrix=Chan_Ip_Matrix'*inv(R);
end
//******************************************************************************
// Get the matched filter.
//******************************************************************************
MF_Coeff=conj([Pre_Amb_IFFT(PRE_AMB_SIZE:-1:1)]);             // Matched filter.
MF_Energy=MF_Coeff*MF_Coeff';                   // Energy of the matched filter.
MF_Coeff=MF_Coeff/MF_Energy;                       // Normalized to unit energy.
//******************************************************************************
// Get the frequency offset matrix.
//******************************************************************************
FOFF_Vec=0:FOFF_VEC_SIZE-1;
FOFF_Vec=FOFF_Vec*%i*FOFF;
FOFF_Vec=exp(FOFF_Vec);
//******************************************************************************
// Frequency offset estimation matrix for step 1.
//******************************************************************************
Freq_Step=2*MAX_FOFF/ML_FOFF1_BINS;
y=[0:ML_FOFF1_BINS]';
y=y*Freq_Step-MAX_FOFF;         // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix=y*x;
FOFF_Est_Matrix=exp(-%i*FOFF_Est_Matrix);
clear x;
//******************************************************************************
// Frequency offset estimation matrix for step 2.
//******************************************************************************
Freq_Step2=2*MAX_FOFF2/ML_FOFF2_BINS;
y2=[0:ML_FOFF2_BINS]';
y2=y2*Freq_Step2-MAX_FOFF2;     // Used in 3D plot of frequency offset estimate.
x=[0:FOFF_EST_RX_SIZE-1];
FOFF_Est_Matrix2=y2*x;
FOFF_Est_Matrix2=exp(-%i*FOFF_Est_Matrix2);
clear x;
//******************************************************************************
// Initialize frequency offset estimation parameters.
//******************************************************************************
PLOT_START_TIME=PRE_AMB_SIZE-EST_CHAN_LEN;
PLOT_END_TIME=PRE_AMB_SIZE+EST_CHAN_LEN;
PLOT_LEN=PLOT_END_TIME-PLOT_START_TIME+1;
//******************************************************************************
// Frequency offset estimation error measurements.
//******************************************************************************
RMS_Err_FOFF1=0.0;         // RMS frequency offset estimation error in 1st step.
Max_Err_FOFF1=-1.0;    // Maximum frequency offset estimation error in 1st step.
RMS_Err_FOFF2=0.0;         // RMS frequency offset estimation error in 2nd step.
Max_Err_FOFF2=-1.0;    // Maximum frequency offset estimation error in 2nd step.
//******************************************************************************
// Arrays for the 3D plot of the frequency offset estimate. 'y' and 'y2'
// arrays already defined above.
//******************************************************************************
x=[PLOT_START_TIME:PLOT_END_TIME];                              // For plotting.
z=zeros(ML_FOFF1_BINS+1,PLOT_LEN);                              // For plotting.
z2=zeros(ML_FOFF2_BINS+1,PLOT_LEN);                             // For plotting.
//******************************************************************************
Ber=0;                                                        // Bit-error-rate.
Erasure=0;                                          // Number of frame erasures.
//******************************************************************************
// Simulation.
//******************************************************************************
for Frame_Cnt=1:MAX_FRAMES
// clf();
//******************************************************************************
// Get the random frequency offset and the frequency offset matrix.
//******************************************************************************
 FOFF=2.0*MAX_FOFF*(grand(1,1,'def')-0.5);
 if abs(FOFF) > MAX_FOFF then
  halt('Freq offset exceeds maximum.');
 end;
 FOFF_Vec=0:FOFF_VEC_SIZE-1;
 FOFF_Vec=FOFF_Vec*%i*FOFF;
 FOFF_Vec=exp(FOFF_Vec);
//******************************************************************************
// Initialize estimates of the frequency offset.
//******************************************************************************
 Est_FOFF1=0.0;
 Est_FOFF2=0.0;
 Est_FOFF3=0.0;
//******************************************************************************
// Generate QPSK symbols.
//******************************************************************************
 [Re_Tx_Sym,Im_Tx_Sym,Tx_Sym]=Get_QPSK_Sym(OFDM_SIZE);
//******************************************************************************
 Sym_IFFT=fft(Tx_Sym.',1).';            // IFFT. There is division by OFDM_SIZE.
//******************************************************************************
// Add cyclic prefix and preamble.
//******************************************************************************
 Chan_Ip=[Pre_Amb_IFFT Sym_IFFT(OFDM_SIZE-EST_CHAN_MEM+1:OFDM_SIZE) Sym_IFFT];
//******************************************************************************
// Get channel.
//******************************************************************************
 [Re_Chan,Im_Chan,Chan]=Get_Channel(CHAN_LEN,FADE_MEAN,FADE_STD);
 Chan_Ext_Zeros=[Chan zeros(1,OFDM_SIZE-CHAN_LEN)];             // Zero padding.
 Chan_FFT=fft(Chan_Ext_Zeros.',-1).';                    // OFDM_SIZE-point FFT.
//******************************************************************************
 Chan_Op=Get_Channel_Op(Chan,Chan_Ip,FOFF_Vec,NOISE_MEAN,NOISE_STD);
 FOFF_Est_Rx=[Chan_Op(1:FOFF_EST_RX_SIZE)];
//******************************************************************************
// ML frequency offset estimation, step 1. 'surf' tested on Scilab 5.3.2.
// 'plot3d' ('surf' not available) tested on Scilab 2.6.
//******************************************************************************
 for k=1:ML_FOFF1_BINS+1
  Demod_Rx=FOFF_Est_Matrix(k,:).*FOFF_Est_Rx;
  FOFF_Sync_Op=convol(MF_Coeff,Demod_Rx);
  z(k,:)=FOFF_Sync_Op(PLOT_START_TIME:PLOT_END_TIME);
 end;
 [Max_Sync_Op,Max_Inx]=max(abs(z));
 Est_FOFF1=(Max_Inx(1)-1)*Freq_Step-MAX_FOFF;
 Freq_Est_Err=Est_FOFF1-FOFF;
 RMS_Err_FOFF1=RMS_Err_FOFF1+Freq_Est_Err*Freq_Est_Err;
 Max_Err_FOFF1=max([Max_Err_FOFF1 abs(Freq_Est_Err)]);
// subplot(211);
// surf(x,y',abs(z).^2,'linestyle','dot','edgecol','red','facecol','inter');
// plot3d(y,log10(x),abs(z).^2,flag=[-3,2,4]);
//******************************************************************************
// ML frequency offset estimation, step 2. 'surf' tested on Scilab 5.3.2.
// 'plot3d' ('surf' not available) tested on Scilab 2.6.
//******************************************************************************
 Demod_Rx=FOFF_Est_Matrix(Max_Inx(1),:).*FOFF_Est_Rx;
 for k=1:ML_FOFF2_BINS+1
  Demod_Rx2=FOFF_Est_Matrix2(k,:).*Demod_Rx;
  FOFF_Sync_Op2=convol(MF_Coeff,Demod_Rx2);
  z2(k,:)=FOFF_Sync_Op2(PLOT_START_TIME:PLOT_END_TIME);
 end;
 [Max_Sync_Op2,Max_Inx2]=max(abs(z2));
 Est_FOFF2=(Max_Inx2(1)-1)*Freq_Step2-MAX_FOFF2;
 Freq_Est_Err=Est_FOFF1+Est_FOFF2-FOFF;
 RMS_Err_FOFF2=RMS_Err_FOFF2+Freq_Est_Err*Freq_Est_Err;
 Max_Err_FOFF2=max([Max_Err_FOFF2 abs(Freq_Est_Err)]);
// subplot(212);
// surf(x,y2',abs(z2).^2,'linestyle','dot','edgecol','red','facecol','inter');
// plot3d(y,log10(x),abs(z2).^2,flag=[-3,2,4]);
//******************************************************************************
// Sync_Op=convol(MF_Coeff,Chan_Op);
// Avg_Sync_Op=sum(abs(Sync_Op))/size(Sync_Op,'c');
// [Peak_Sync_Op,Peak_Inx]=max(abs(Sync_Op));
// PAR(Frame_Cnt)=Peak_Sync_Op/Avg_Sync_Op;
// subplot(221);
// plot(abs(Sync_Op).^2);
// Rx_Sig=Chan_Op(EST_CHAN_LEN:FRAME_SIZE);             // Steady-state samples.
//******************************************************************************
// Demodulation using the final estimate of the frequency offset.
//******************************************************************************
 Est_FOFF_Vec=0:FOFF_VEC_SIZE-1;
 Est_FOFF_Vec=Est_FOFF_Vec*%i*(Est_FOFF1+Est_FOFF2);
 Est_FOFF_Vec=exp(-Est_FOFF_Vec);
 Chan_Op=Chan_Op.*Est_FOFF_Vec;
//******************************************************************************
// Frame detection in the absence of frequency offset.
//******************************************************************************
// if IDEAL_CHEST == 0 then
//  Diff_Inx=Peak_Inx-PRE_AMB_SIZE;
// else
//  Diff_Inx=0;
// end
//******************************************************************************
// Frame detection in the presence of frequency offset.
//******************************************************************************
 if IDEAL_CHEST == 0 then
  Diff_Inx=Max_Inx(2)-EST_CHAN_LEN-1;
 else
  Diff_Inx=0;
 end
//******************************************************************************
 if ((Diff_Inx >= 0) & (Diff_Inx < CHAN_LEN)) then
  Start_Inx1=EST_CHAN_LEN-Diff_Inx;
  End_Inx1=PRE_AMB_SIZE-Diff_Inx;
  Rx_Sig1=Chan_Op(Start_Inx1:End_Inx1);
  if MMSE_CHEST == 0 then
   Chan_Est=Chan_Est_Matrix*Rx_Sig1.';
  elseif MMSE_CHEST == 1 then
   Chan_Est=MMSE_Chan_Est_Matrix*Rx_Sig1.';
  end
  Chan_Est_Ext_Zeros=[Chan_Est;zeros(OFDM_SIZE-EST_CHAN_LEN,1)];
  Est_Chan_FFT=fft(Chan_Est_Ext_Zeros,-1).';
//  subplot(222);
//  plot2d3(abs(Chan_Est));
//  subplot(224);
//  plot2d3(abs(Chan));
//  subplot(211);
//  plot(abs(Chan_FFT));
//  subplot(212);
//  plot(abs(Est_Chan_FFT))
  Start_Inx2=PRE_AMB_SIZE+EST_CHAN_LEN-Diff_Inx;
  End_Inx2=FRAME_SIZE-Diff_Inx;
  Rx_Sig2=Chan_Op(Start_Inx2:End_Inx2);                              // For FFT.
  Rx_FFT_Op=fft(Rx_Sig2.',-1).';                     // FFT of received samples.
  if IDEAL_CHEST == 0 then
   Dist1=abs(Rx_FFT_Op-QPSK_SYM1.*Est_Chan_FFT);         // Sq. dist. wrt sym 1.
   Dist2=abs(Rx_FFT_Op-QPSK_SYM2.*Est_Chan_FFT);         // Sq. dist. wrt sym 2.
   Dist3=abs(Rx_FFT_Op-QPSK_SYM3.*Est_Chan_FFT);         // Sq. dist. wrt sym 3.
   Dist4=abs(Rx_FFT_Op-QPSK_SYM4.*Est_Chan_FFT);         // Sq. dist. wrt sym 4.
  else
   Dist1=abs(Rx_FFT_Op-QPSK_SYM1.*Chan_FFT);             // Sq. dist. wrt sym 1.
   Dist2=abs(Rx_FFT_Op-QPSK_SYM2.*Chan_FFT);             // Sq. dist. wrt sym 2.
   Dist3=abs(Rx_FFT_Op-QPSK_SYM3.*Chan_FFT);             // Sq. dist. wrt sym 3.
   Dist4=abs(Rx_FFT_Op-QPSK_SYM4.*Chan_FFT);             // Sq. dist. wrt sym 4.
  end
  [M,K]=min(Dist1,Dist2,Dist3,Dist4);              // Get min. dist and indices.
  Rx_Sym=round(exp(%i*(K*%pi/2-%pi/4)));                   // Estimated symbols.
  Ber=Ber+nnz(Tx_Sym-Rx_Sym);
 else
  Erasure=Erasure+1;                                 // Erase the current frame.
//  pause;
 end
end
//******************************************************************************
RMS_Err_FOFF1=sqrt(RMS_Err_FOFF1/MAX_FRAMES);
RMS_Err_FOFF2=sqrt(RMS_Err_FOFF2/MAX_FRAMES);
timer()
//******************************************************************************
